/* Module init_ki.c
*  Initializes all (or most) of the PoCa specific stuff.
*/

#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <math.h>

#include <at200.h>
#include <camera.h>

#include "idecls.h"
#include "bitmap.h"
#include "camsys.h"
#include "proto.h"
#include "mode.h"


char	line[ MAXLINE + 1 ];
int	lastlogmin, lastlogsec,	loginterval, lastsamplemin, lastsamplesec;

extern void s_cleanup( void );

/* Display usage of KI */

void show_usage( void )
{
  printf("\nUsage: KI [-Start filename1] [-Xeq filename2]\n\n"
  	 "  where: '-Start filename1	Name of KI start/config file.\n"
	 "         '-Xeq   filename2    Name of script file to execute.\n" );
  exit ( 1 );
}

void _cdecl init_poca( int argc, char *argv[] )
{
	register int len;
	register char *ep;
	entry_ptr pv[50];
	int i, next;
	char	startfile[PATHSIZE], startflname[50];
	pv_ptr p = (void *)NULL;
		
	_dos_getdate( &dt );
	std0_misc.dt = std1_misc.dt = dt;
	
	_dos_gettime( &tm );
	std0_misc.tm = std1_misc.tm = tm;
	
	textmode = TRUE;	/* Set default video mode to standard text. */
	sampleptr = no_samples = 0;
	do_compress[0] = FALSE;	/* Set default to no compression */
	do_compress[1] = FALSE;
	init_ki();
	genbitmap();

	lastlogmin = tm.minute;
	lastlogsec = tm.second;
	lastsamplemin = tm.minute;
	lastsamplesec = tm.second;
	qlookmode[0] = qlookmode[1] = cmndlogmode = templogmode = dispmode = 0;
	inhibit0 = inhibit1 = inhibit_display = 0;
	sampleinterval = 15;
	loginterval    = 120;
	line[0] = startflname[0] = '\0';
	
	for ( i=1; i<argc; i++ )	/* Process command line args */
	{
		if ( argv[i][0] == '-' )
			switch( argv[i][1] )
			{
			case 's':	/* Get startup file name. */ 	
			case 'S':
				   if ( ++i > argc )
				   	show_usage();
				   strcpy ( startflname, argv[ i ] );
				   break;
			case 'x':			/* XEQ <filename> */
			case 'X':
				   if ( ++i > argc )
				   	show_usage();
				   sprintf ( line, "xeq %s", argv[ i ] );
				   break;
			default:   show_usage();
			}
		else
			show_usage();
	}

	if ( !startflname[0] )		/* Define full path to startup file. */
	{
		strcpy ( startfile, filespath );
		strcat ( startfile, "KI_START.DAT" );
	}
	else
	{
		strcpy ( startfile, filespath );
		strcat ( startfile, startflname );
	}

	do_startup( startfile );	/* Process startup file. */
	s_com_init();	 	/* Initialize AFB communications */
	select_camera( 1 );  p = reset_afb( p );
	select_camera( 0 );  p = reset_afb( p );

	if ( cmndlogmode )
	{
		fprintf( cmndlogfl, "KI started: %2d/%2d/%2d  %2d:%2d\n", 
			dt.year, dt.month, dt.day, tm.hour, tm.minute );
		fflush( cmndlogfl );
	}
	if ( line[0] )	/* True if there was a "-x scr_file" on the command line */
	{
	   if (( ep = iparse(line, pv, &next)) != NULL )
	      printf( "%s", ep );
	   else
	   {
	      pv[next] = NULL;
	      iexec(pv);
	   }
	}
        line[0] = '\0';
}



void init_ki ( void )	/* Initialize camera structures. */
{
  char *env, fname[80];
  unsigned int count;
  
  if ( !emm_present() ) 	 /* Check for EMM capability. */
  {
  	setvideomode( TEXT_MODE );
  	printf( "EMM capability not present - check system configuration.\n" );
  	exit( -1 );
  } 

  if ( emm_version() < 0x40 )
  {
  	setvideomode( TEXT_MODE );
  	printf( "EMM version incorrect - check system configuration.\n" );
  	exit( -1 );
  } 

  ems_data = (unsigned *) ems_get_page_frame();
  if ( !ems_data )
  {
  	setvideomode( TEXT_MODE );
  	printf( "Unable to locate EMS page frame - check system configuration.\n" );
  	exit( -1 );
  }
  std0_misc.cam_opmode = 3;		/* Both cameras enabled. */
  std1_misc.cam_opmode = 3;
  
  select_camera( 1 );     /* Initialize camera 1 structures. */
  init_struct();

  select_camera( 0 );      /* then camera 0 */
  init_struct();

  /* Get environment variables for path names. */

  if ( ( env = getenv( "KI_LOG" ) ) != (char *) NULL )
  {
     strcpy( templogflname, env );	/* Build temperature log file path. */
     strcat( templogflname, "\\" );
     strcpy( cmndlogflname, env );	/* Build command log file path. */
     strcat( cmndlogflname, "\\" );
	 strcpy( qlookpathname, env );	// Quiklook file path.
	 strcat( qlookpathname, "\\" );
  }
  else
  {
     templogflname[0] = '\0';		/* No log file path specified. */
     cmndlogflname[0] = '\0';
	 qlookpathname[0] = '\0';
  }

  if ( ( env = getenv( "KI_IMAGE" ) ) != (char *) NULL )
  {
     strcpy( imagepath, env );		/* Base path for images. */
     strcat( imagepath, "\\" );
  }
  else
     imagepath[0] = '\0';

  if ( ( env = getenv( "KI_SCRIPT" ) ) != (char *) NULL )
  {
     strcpy( scriptpath, env );		/* Base path for XEQ script files. */
     strcat( scriptpath, "\\" );
  }
  else
     scriptpath[0] = '\0';

  if ( ( env = getenv( "KI_FILES" ) ) != (char *) NULL )
  {
     strcpy( filespath, env );		/* Base path for start-up files. */
     strcat( filespath, "\\" );
  }
  else
     filespath[0] = '\0';

  /* Open KI sequence data file. */

  strcpy ( fname, filespath );
  strcat ( fname, "KI_DATA.DAT" );
  fdata = open( fname, O_RDWR | O_CREAT | O_BINARY , S_IREAD | S_IWRITE );
  if ( fdata < 0 )
  {
     	sprintf ( omsg, "Unable to open file %s. (disk full ?)\n", fname );
     	disp_omsg( ERROR );
     	exit( -1 );
  }

  /* Get the file sequence numbers. */

  count = read ( fdata, &ki_data, sizeof ( KI_DATA ) );
  if ( ( count < 1 ) || ( ki_data.dt.day != dt.day ) )	/* Empty file ? */
  {
     memset( &ki_data, '\0', sizeof ( KI_DATA ) );	/* reset to zeros */
     ki_data.dt = dt;					/* Set effective date */
  }
}



void terminate_ki ( void )	/* Program termination. */
{
	if ( cmndlogmode )
	{
		_dos_getdate( &dt );
		_dos_gettime( &tm );
		fprintf( cmndlogfl, "KI ended: %2d/%2d/%2d  %2d:%2d\n", 
			dt.year, dt.month, dt.day, tm.hour, tm.minute );
		fclose( cmndlogfl );
	}
	if ( templogmode )
		close( templogfl );
	close( fdata );		/* Close file sequence data. */

    ems_free( std0_misc.ems_handle );	/* Release EMS memory. */
    ems_free( std1_misc.ems_handle );
    s_cleanup();
    setvideomode( TEXT_MODE );
	exit( -1 );
}



void init_struct ( void )
{
  unsigned int i, j, x;
  
  /* Allocate EMS memory for this structure. */   
     
  i = ems_alloc( 524288L );	/* Maximum image size 512X512 X 16 bits. */

  if( !i ) 
  {
  	setvideomode( TEXT_MODE );
	printf( "Unable to allocate EMS memory. Cannot proceed.\n");
	exit( -1 );
  }

  sel_misc->ems_handle = i;
  sel_misc->image_size = 0L;

  sel_sys->ccd = ccd[4];	/* CCD type - TH7985B 512 X 512 pixels */
  if ( sel_cam == 0 )
  {
     sel_sys->at200_id = 2;	/* Controller ID and base addr */
     at200_base = 0x220;
  }
  else
  {
     sel_sys->at200_id = 3;
     at200_base = 0x230;
  }
  sel_sys->frame = 0;		/* This is a full frame CCD */
  sel_sys->open_delay = 5;	
  sel_sys->close_delay = 5;
  sel_sys->preamp_delay = 7;
  sel_sys->speed = A200KHZ;	/* 200K Hz Analog to digital convertor */
  sel_sys->data_bits = 14;	/* with 14 bit/pixel resolution */
      
  sel_exp->exposure = 200;	/* default exposure */
  sel_exp->gain = 0;		/* 0 = 1X, 1 = 4X gain */
  sel_exp->mpp_mode = 1;	/* allow/disallow MPP exposures */
  sel_exp->corg = 0;		/* sub-array col origin, row origin */
  sel_exp->rorg = 0;
  sel_exp->cols = 512;		/* subarray # of rows, cols */
  sel_exp->rows = 512;
  sel_exp->pbin = 2;		/* Column binning factor */
  sel_exp->sbin = 2;		/* Row binning factor */
  sel_exp->clr_cnt = 5;		/* # of times to clear CCD before exposure */
  sel_exp->row_shift = 0;

  sel_misc->camera = (unsigned char) sel_cam;	/* Camera number. (constant) */
  sel_misc->fw = 0;		/* Filter number. (supplied by AFB) */
  sel_misc->filter_desc[0] = '\0';	/* Filter description */
  sel_misc->preamp = 0;		/* Preamp on/off during exposure. */
  sel_misc->exp_scale = 1;	/* Exposure scale 1=ms, 1000 = sec, 60000=min */
  for ( x=0; x<5; x++ )	 	/* Temperature readings from AFB */
  {
     sel_misc->temp.tei[x] = 0;
     sel_misc->temp.tee[x] = 0;
     sel_misc->temp.td[x]  = 0;
     sel_misc->temp.tcc[x] = 0;
     sel_misc->temp.tcf[x] = 0;
     sel_misc->temp.tfw[x] = 0;
  }
  sel_misc->sequence = 0;		/* file sequence # */
  sel_misc->exposure_type = 'A';	/* Airglow/Aurora, Bias, Dark, Flat */
  sel_misc->version = (1 * 16) + 1;	/* file/program version = 1.10 */
  strcpy ( sel_misc->jpeg_qual, "   " );	/* JPEG quality */
  strcpy ( sel_misc->comp_fact, "1.0" );	/* Compression factor */
  strcpy ( sel_misc->os_version, "5.0" );	/* Operating System version */
  sel_misc->lens_desc[0] = '\0'; 
  sel_misc->filter_desc[0] = '\0'; 
  sel_misc->location[0] = '\0'; 
  sel_misc->azimuth[0] = '\0'; 
  sel_misc->operator[0] = '\0'; 
  
  dsp_command( DSP_HALT );		/* Stop any operation in progress. */

	  /* Load setup program into camera */
  x = load_dsp_program( dsp_cam200k, size_of_dsp_cam200k );

  if( !x ) 
  {
    setvideomode( TEXT_MODE );
    printf("ERROR: Unable to load program into AT200 at addr %X\n", at200_base );
    exit( 1 );
  }

  /* Wait for DSP to complete initialization */
  
  for( x=1; x; x++ )
    if( !(inp( at200_base+DSP_INTERRUPT_STATUS) & (HOST_FLAG_2 | HOST_FLAG_3)) )
      break;  
  if( !x ) 
  {  
    setvideomode( TEXT_MODE );
    printf("AT200 at addr %X is unresponsive or missing.  Unable to proceed.\n",
           at200_base );
    exit( 2 );
  }

  /* Reset filter descriptions. */
  
  for ( i=0; i<2; i++ )
     for ( j=0; j<5; j++ )
        filter_desc[i][j][0] = '\0';
        
  /* Clear AFB data conversion tables. */

  memset( &afb_const[0], '\0', sizeof ( AFB_CONST ) );
  memset( &afb_const[1], '\0', sizeof ( AFB_CONST ) );
  
  memset( t_history0, 0, sizeof( t_history0 ) );
  memset( t_history1, 0, sizeof( t_history1 ) );
        
}

/* Display message to user, and write to log if enabled. */

void disp_omsg( int errormess )
{
int mode;
	/* Inhibit display prevents output to screen during
	 * temperature logging. 
	 */
	if ( (errormess  == NOERROR) && (textmode == TRUE) ){
		printf("%s", omsg);
		if ( cmndlogmode ){
			fprintf( cmndlogfl, "%s", omsg );
			fflush( cmndlogfl );
		}
		return;
	}
	
	if ( errormess == ERROR ){	 
		mode = getvideomode();
		if ( mode != TEXT_MODE )
			setvideomode( TEXT_MODE );
		if ( !inhibit_display )
			printf( "%s", omsg );
	}
}



/* Process KI startup file. 
 *
 * Input is in the form: -keyword [keyvalue] [keyvalue]
 * where keywords may have zero or more keyvalues to use.
 */

void do_startup ( char *startfilename )
{
char	*token;
char	fname[13];
int	v1, v2, done, get_next = 1;


	if ( !set_cmdsrc ( startfilename, CMD_FILE ) )
	{
		printf( "ERROR: Unable to open startup file: %s\n",
			startfilename );
		terminate_ki();
	}
	
	while ( 1 )
	{
	    if ( get_next )
		token = get_nxt_tok();
	    if ( token == (void *) NULL )
	    	break;
	    else if ( token[0] == '-' )		/* Keyword */
	    {
		if ( strmatch( &token[1], "temperature_log" ) )
		{
			templogmode = 1;	/* Temperature logging */

			/* See if file name specified. */
				   
			token = get_nxt_tok();

			if( token[0] == '-')
			{
				strcat ( templogflname, "KI_TEMP.LOG" );
				get_next = 0;
			}
			else 
			{
				strcat ( templogflname, token );
				get_next = 1;
			}
			templogfl = open( templogflname, 
					O_CREAT | O_WRONLY | O_APPEND | O_BINARY,
					S_IREAD | S_IWRITE );
			duptlogfl = dup( templogfl );
		}
		else if ( strmatch ( &token[1], "log" ) )
		{
			cmndlogmode = 1;	/* Log commands */
			token = get_nxt_tok();

			if( token[0] == '-')
			{
				strcat ( cmndlogflname, "KI_CMND.LOG" );
				get_next = 0;
			}
			else 
			{
				strcat ( cmndlogflname, token );
				get_next = 1;
			}
			cmndlogfl = fopen ( cmndlogflname, "a" );
		}
		else if ( strmatch ( &token[1], "qlook" ) ) //Generate quick look file?
		{
			qlookmode[0] = get_nxt_number();
			qlookmode[1] = get_nxt_number();
			if ( (qlookmode[0] != 0) || (qlookmode[1] != 0) ){
				sprintf( fname, "QLOOK.%02d", dt.day );
				strcat ( qlookpathname, fname );
				qlookfptr = fopen ( qlookpathname, "a" );
				fclose( qlookfptr );	// Just create the file now.
			}
			get_next = 1;
		}
		else if ( strmatch ( &token[1], "display" ) )
		{
			dispmode = 1;	/* Display cmnds */
			get_next = 1;	/* from script files */
		}
		else if ( strmatch ( &token[1], "interval" ) )
		{
			sampleinterval = get_nxt_number();
			loginterval = get_nxt_number();
			get_next = 1;
		}
		else if ( strmatch ( &token[1], "location" ) )
		{
			token = get_nxt_tok();
			if ( strlen ( token ) > 80 )
				token[80] = '\0';
			strcpy ( std0_misc.location, token );
			strcpy ( std1_misc.location, token );
		}

		/* Process '-camera n m' line where:
		 * n = camera number (0 or 1)
		 * m = AFB serial port address for camera ( 1 thru 4)
		 */ 

		else if ( strmatch ( &token[1], "camera" ) )
		{
			v1 = get_nxt_number();		/* Camera # 0 or 1 */
			if ( !( (v1 == 0) || (v1 == 1) ) )
			{
				printf( "ERROR: Camera number must be 0 or 1\n");
				terminate_ki();
			}

			select_camera( v1 );

			v2 = get_nxt_number();		/* AFB serial portno */
			if ( !( (v2 == 1) || (v2 == 2) || (v2 == 3) || 
				(v2 == 4) ) )
			{
				printf("ERROR: AFB port number must be 1 thru 4\n");
				terminate_ki();
			}
			sel_misc->portno = (unsigned char) v2;

			done = 0;
			while ( !done )
			{
				token = get_nxt_tok();
				if ( strmatch ( &token[1], "filter" ) )
				{
				    v2 = get_nxt_number();	/* Filter # */
				    if ((v2 >= 0) && (v2 < 5 ))
				    {
					token = get_nxt_tok();
					if ( strlen ( token ) > 5 )
						token[5] = '\0';
					strcpy ( &filter_desc[v1][v2][0], token );
				    }
				    else
				    {
					printf("ERROR: Filter number must be 0 thru 4\n");
					terminate_ki();
				    }
				}
				else if ( strmatch ( &token[1], "lens" ) )
				{
					token = get_nxt_tok();
					if ( strlen ( token ) > 5 )
						token[5] = '\0';
					strcpy ( sel_misc->lens_desc, token );
				}
				else if ( strmatch ( &token[1], "azimuth" ) )
				{
					token = get_nxt_tok();
					if ( strlen ( token ) > 10 )
						token[10] = '\0';
					strcpy ( sel_misc->azimuth, token );
				}
				else if ( strmatch ( &token[1], "operator" ) )
				{
					token = get_nxt_tok();
					if ( strlen ( token ) > 3 )
						token[3] = '\0';
					strcpy ( sel_misc->operator, token );
				}
				else
					done = 1;
			}
			get_next = 0;
		}
		
		/* Process -AFB x y  :  x = AFB number, y = AFB ID */
 		 
		else if ( strmatch ( &token[1], "afb" ) )
		{
			v1 = get_nxt_number();			/* AFB # */
			if ( !( (v1 == 0) || (v1 == 1) ) )
			{
				printf("ERROR: AFB number must be 0 or 1\n");
				terminate_ki();
			}

			afb_const[v1].id = get_nxt_number();	/* AFB ID */

			/* Process all AFB data conversion setups. */

			done = 0;
			while ( !done )
			{
				token = get_nxt_tok();
				if ( strmatch ( &token[1], "tei" ) )
				{
				    afb_const[v1].tei[0] = get_nxt_number();
				    afb_const[v1].tei[1] = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "tee" ) )
				{
				    afb_const[v1].tee[0] = get_nxt_number();
				    afb_const[v1].tee[1] = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "td" ) )
				{
				    afb_const[v1].td[0] = get_nxt_number();
				    afb_const[v1].td[1] = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "tcf" ) )
				{
				    afb_const[v1].tcf[0] = get_nxt_number();
				    afb_const[v1].tcf[1] = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "tfw" ) )
				{
				    afb_const[v1].tfw[0] = get_nxt_number();
				    afb_const[v1].tfw[1] = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "plus5" ) )
				{
				    afb_const[v1].plus5 = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "plus15" ) )
				{
				    afb_const[v1].plus15 = get_nxt_number();
				}
				else if ( strmatch ( &token[1], "minus15" ) )
				{
				    afb_const[v1].minus15 = get_nxt_number();
				}
				else
					done = 1;
			}
			get_next = 0;
		}
		else if ( strmatch ( &token[1], "stop" ) )
		{
			token = get_nxt_tok();
			inhibit0 = atoi ( token );

			token = get_nxt_tok();
			inhibit1 = atoi ( token );
			get_next = 1;
		}
		else if ( token == (void *) NULL)
			break;
		else
		{
			printf ( "ERROR: Unknown keyword '%s' in startup file '%s'\n",
				token,  startfilename );
			terminate_ki();
		}
	    }
	    else
	    {
		printf ( "ERROR: Unknown keyword '%s' in startup file '%s'\n",
			token,  startfilename );
		terminate_ki();
	    }
	}
}


/* Compare first 'n' chars of in_token to target.
 * Return true if they match, false otherwise.  */
  
int strmatch ( char *in_token, char *target )
{
register int i, j;

	j = strlen ( in_token );
	if ( j )
	{
		for ( i=0; i<j; i++ )
		{
			if ( !target[ i ] )
				return 0;   /* in_token longer than target */
			if ( !( target[ i ] == in_token[ i ] ) )
				return 0;   /* No compare */
		}
	}
	else
		return 0;		    /* NULL string */
		
	return 1;
}

/* Get next token from startup command file or command line.
 * Return pointer to stringified token, or NULL if EOF.
 */
 
char *get_nxt_tok( void )
{
static char	outstring[80];
int		i, stopper;

	if ( cmdsrc == CMD_LINE )
	{
		if(( cmdln[cmdlnx] == '\0' ) || ( cmdln[cmdlnx] == '\n' ))
		{
			cmdsrc = 0;
			return( (void *) NULL );
		}
	}
	else if ( cmdsrc == CMD_FILE )
	{
		/* Get a line of input from the startup file. */

		while ( cmdln[cmdlnx] == '\n' )
		{
			if ( fgets ( cmdln, 100, f1 ) == (void *) NULL )
			{
				fclose ( f1 );		/* EOF */
				cmdsrc = 0;
				return( (void *) NULL );
			}
			cmdlnx = 0;
			strcpy ( omsg, cmdln );
			disp_omsg( NOERROR );
		}
	}
	else
		return( (void *) NULL );
	
	/* Skip over white space to next token. */
	
	while (	(cmdln[cmdlnx] == ' ') || (cmdln[cmdlnx] == '\t') )
		cmdlnx++;

	/* Copy next token to output string. */

	i = 0;
	if ( cmdln[cmdlnx] == '"' )
	{
		stopper = '"';
		cmdlnx++;
	}
	else
		stopper = ' ';
		
	while ( 1 )
	{
		if (	(cmdln[cmdlnx] == (char) stopper ) 
		     || (cmdln[cmdlnx] == '\t' ) 
		     || (cmdln[cmdlnx] == '\n' ) )
		     break;
		else
			outstring[ i++ ] = cmdln[ cmdlnx++ ];
		if ( i == 80 )
			break;
	}

	outstring[ i ] = '\0';
	return strlwr ( outstring );
}


/* Get next token from startup file and return as numeric integer value. 
 * Set the source of command line args for above two functions.
 */

int get_nxt_number ( void )
{
char	*token;
int	i, j;

	token = get_nxt_tok();
	j = strlen ( token );
	for ( i=0; i<j; i++ )
	if ( !((token[i] == '-') || ((token[i] >= '0') && (token[i] <= '9' ))))
	{
		printf("ERROR: Numeric value expected.\n");
		terminate_ki();
 	}
	return ( atoi ( token ) );	
}

/* Set source of command args to file, or buffer. */

int set_cmdsrc ( char *source, int type )
{
	cmdlnx = 0;
	cmdsrc = 0;
	if ( type == CMD_FILE )
	{
		if ( ( f1 = fopen ( source, "r" ) ) == (FILE *) NULL )
			return ( 0 );
		cmdln[ cmdlnx ] = '\n';	/* to force read from file */
	}
	else if ( type == CMD_LINE )
		strcpy ( cmdln, source );
	else
		return ( 0 );
	cmdsrc = type;
	return ( 1 );		/* success */
}



void genbitmap( void )
{
int i, x, y;
int rows, cols, radius, maxradius, bmapsize;

rows = sel_exp->rows/sel_exp->pbin;
cols = sel_exp->cols/sel_exp->sbin;
maxradius = cols/2;

i = rows * cols;
bmapsize = (i >> 8)*sizeof( char );
bitmap = (unsigned char *)(malloc( bmapsize ));
memset( bitmap, 0, bmapsize );

while ( i-- )
{
	x = i % cols; y = i / rows;
	radius = (int)(sqrt( (double)( x*x + y*y ) ));
	if ( radius < maxradius )
		setbit( bitmap, i );
}
}

